/*

------------------------------------------------------------------------------

A license is hereby granted to reproduce this software source code and
to create executable versions from this source code for personal,
non-commercial use.  The copyright notice included with the software
must be maintained in all copies produced.

THIS PROGRAM IS PROVIDED "AS IS". THE AUTHOR PROVIDES NO WARRANTIES
WHATSOEVER, EXPRESSED OR IMPLIED, INCLUDING WARRANTIES OF
MERCHANTABILITY, TITLE, OR FITNESS FOR ANY PARTICULAR PURPOSE.  THE
AUTHOR DOES NOT WARRANT THAT USE OF THIS PROGRAM DOES NOT INFRINGE THE
INTELLECTUAL PROPERTY RIGHTS OF ANY THIRD PARTY IN ANY COUNTRY.

Copyright (c) 1994-2014, John Conover, All Rights Reserved.

Comments and/or bug reports should be addressed to:

    john@email.johncon.com (John Conover)

------------------------------------------------------------------------------

tsinvestsim-lognormal.c for summing the individual values of a time interval,
to obtain the median of the sum, and the mean of the sum, in of a file
in tsinvestsim(1) format.

The input file structure is a text file consisting of records, in
temporal order, one record per time series sample.  Blank records are
ignored, and comment records are signified by a '#' character as the
first non white space character in the record. Data records must
contain at least one field, which is the data value of the sample, but
may contain many fields-if the record contains many fields, then the
first field is regarded as the sample's time, and the last field as
the sample's value at that time.

$Revision: 0.0 $
$Date: 2006/01/18 19:36:00 $
$Id: tsinvestsim-lognormal.c,v 0.0 2006/01/18 19:36:00 john Exp $
$Log: tsinvestsim-lognormal.c,v $
Revision 0.0  2006/01/18 19:36:00  john
Initial version


*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <unistd.h>

#ifdef __STDC__

#include <float.h>

#endif

static char rcsid[] = "$Id: tsinvestsim-lognormal.c,v 0.0 2006/01/18 19:36:00 john Exp $"; /* program version */
static char copyright[] = "Copyright (c) 1994-2006, John Conover, All Rights Reserved"; /* the copyright banner */

#define BUFLEN BUFSIZ /* i/o buffer size */

#define TOKEN_SEPARATORS " \t\n\r\b," /* file record field separators */

#define ARRAY_SIZE 100000 /* array size for sorting values in an interval to find the median in the interval */

#ifdef __STDC__

static const char *help_message[] = /* help message index array */

#else

static char *help_message[] = /* help message index array */

#endif

{
    "\n",
    "Take the sum of a time series in tsinvestsim(1) format\n",
    "Usage: tsinvestsim-lognormal [-d] [-v] [filename]\n",
    "    -d, dump the last interval sorted data values\n",
    "    -v, print the program's version information\n",
    "    filename, input filename\n"
};

#ifdef __STDC__

static const char *error_message[] = /* error message index array */

#else

static char *error_message[] = /* error message index array */

#endif

{
    "No error\n",
    "Error in program argument(s)\n",
    "Error opening file\n",
    "Error closing file\n",
    "Error array size exceeded\n",
    "Error interval array error\n"
};

#define NOERROR 0 /* error values, one for each index in the error message array */
#define EARGS 1
#define EOPEN 2
#define ECLOSE 3
#define ASIZE 4
#define EARRAY 5

#define MIDWAY(a,b) (int) (((unsigned int) (a) + (unsigned int) (b)) / (unsigned int) 2)

#ifdef __STDC__

static void print_message (int retval); /* print any error messages */
static int strtoken (char *string, char *parse_array, char **parse, char *delim); /* parse a record based on sequential delimiters */
static void double_sort (double *array, int bottom, int top); /* binary sort routine for type double */

#else

static void print_message (); /* print any error messages */
static int strtoken ();  /* parse a record based on sequential delimiters */
static void double_sort (); /* binary sort routine for type double */

#endif

#ifdef __STDC__

int main (int argc, char *argv[])

#else

int main (argc, argv)
int argc;
char *argv[];

#endif

{
    char buffer[BUFLEN], /* i/o buffer */
         parsebuffer[BUFLEN], /* parsed i/o buffer */
         lasttoken[BUFLEN], /* last time token */
         *token[BUFLEN / 2], /* reference to tokens in parsed i/o buffer */
         token_separators[] = TOKEN_SEPARATORS;

    int count = 0, /* input file record counter */
        interval_data_count = 0, /* current interval data count */
        retval = NOERROR, /* return value, assume no error */
        fields, /* number of fields in a record */
        dflag = 0, /* dump the last interval sorted data values flag */
        c; /* command line switch */

    double sum = (double) 0.0, /* running value of cumulative sum */
           currentvalue, /* value of current sample in time series */
           interval_data[ARRAY_SIZE], /* values in current interval */
           median_interval_data = (double) 0.0; /* the median of the values in current interval */

    FILE *infile = stdin; /* reference to input file */

    while ((c = getopt (argc, argv, "dv")) != EOF) /* for each command line switch */
    {

        switch (c) /* which switch? */
        {

            case 'd':

                dflag = 1; /* set the dump the last interval sorted data values flag */
                break;

            case 'v':

                (void) printf ("%s\n", rcsid); /* print the version */
                (void) printf ("%s\n", copyright); /* print the copyright */
                optind = argc; /* force argument error */
                retval = EARGS; /* assume not enough arguments */

            default: /* illegal switch? */

                optind = argc; /* force argument error */
                retval = EARGS; /* assume not enough arguments */
                break;
        }

    }

    lasttoken[0] = '\0'; /* null the last time token */

    if (retval == NOERROR) /* enough arguments? */
    {
        retval = EOPEN; /* assume error opening file */

        if ((infile = argc <= optind ? stdin : fopen (argv[optind], "r")) != (FILE *) 0) /* yes, open the input file */
        {
            retval = NOERROR; /* assume no error */

            while (retval == NOERROR && fgets (buffer, BUFLEN, infile) != (char *) 0) /* read the records from the input file */
            {

                if ((fields = strtoken (buffer, parsebuffer, token, token_separators)) != 0) /* parse the record into fields, skip the record if there are no fields */
                {

                    if (token[0][0] != '#') /* if the first character of the first field is a '#' character, skip it */
                    {

                        if (strcmp (lasttoken, token[0]) != 0) /* record from same time interval? */
                        {

                            if (count != 0) /* first record? */
                            {

                                if (interval_data_count > 1) /* enough data in current interval? */
                                {
                                    (void) double_sort (interval_data, 0, interval_data_count - 1); /* sort the values in current interval */
                                    median_interval_data = interval_data[(unsigned int) floor (interval_data_count / 2) - 1]; /* save the median of the values in current interval */
                                }

                                else
                                {
                                    retval = EARRAY; /* array size exceeded */
                                    break; /* stop */
                                }

                                (void) printf ("%s\t%f\t%f\t%f\t%f\n", lasttoken, interval_data[0], median_interval_data, sum / (double) interval_data_count, interval_data[interval_data_count - 1]); /* print the sample's time, minimum, median, mean, and maximum */
                            }

                            (void) strcpy (lasttoken, token[0]); /* save the time interval */
                            sum = (double) 0.0; /* null the cumulative sum of the time series */
                            interval_data_count = 0; /* reset the current interval data count */
                        }

                        currentvalue = atof (token[fields - 1]); /* save the value of the current sample in the time series */

                        if (interval_data_count < ARRAY_SIZE) /* array size for sorting values in an interval to find the median in the interval, exceeded? */
                        {
                            interval_data[interval_data_count] = currentvalue, /* save the values in current interval */
                            interval_data_count ++; /* increment the current interval data count */
                        }

                        else
                        {
                            retval = ASIZE; /* array size exceeded */
                            break; /* stop */
                        }

                        sum = sum + currentvalue; /* add the value  of the current sample in the time series to the cumulative sum of the time series */
                        count ++; /* increment the count of records from the input file */
                    }

                }

            }

            if (count != 0) /* first record? */
            {

                if (interval_data_count > 1) /* enough data in current interval? */
                {
                    (void) double_sort (interval_data, 0, interval_data_count - 1); /* sort the values in current interval */
                    median_interval_data = interval_data[(unsigned int) floor (interval_data_count / 2) - 1]; /* save the median of the values in current interval */
                }

                else
                {
                    retval = EARRAY; /* array size exceeded */
                }

                (void) printf ("%s\t%f\t%f\t%f\t%f\n", lasttoken, interval_data[0], median_interval_data, sum / (double) interval_data_count, interval_data[interval_data_count - 1]); /* print the sample's time, minimum, median, mean, and maximum */

                if (dflag == 1) /* dump the last interval sorted data values flag set? */
                {

                    for (count = 0; count < interval_data_count; count ++)
                    {
                        (void) printf ("# %f\n", interval_data[count]);
                    }

                }

            }

            if (argc > optind) /* using stdin as input? */
            {

                if (fclose (infile) == EOF) /* no, close the input file */
                {
                    retval = ECLOSE; /* error closing file */
                }

            }

        }

    }

    print_message (retval); /* print any error messages */
    exit (retval); /* exit with the error value */

#ifdef LINT

    return (0); /* for lint formalities */

#endif

}

/*

Print any error messages.

static void print_message (int retval);

I) Data structures:

    A) The help_message array is an array of character stings, one per
       line to be printed for help.

    B) The error_message array is an array of character strings, one
       line per error; the error_message array is implicitly addressed
       by the value integer retval, which specifies the error to be
       printed.

II) Function execution:

    A) Depending on the value of the integer retval:

        1) If retval is zero, print nothing-normal/successful program
           exit.

        2) Else if retval is unity, print help.

        3) Else retval is not zero or unity, it is an error code, print
           the corresponding error message.

Returns nothing.

*/

#ifdef __STDC__

static void print_message (int retval)

#else

static void print_message (retval)
int retval;

#endif

{
    size_t help_ctr; /* help_message line counter */

    switch (retval) /* which return value */
    {

        case 0: /* program ended without errors, print nothing */

            break;

        case 1: /* program ended with a request for help, or argument error, print help */

            for (help_ctr = 0; help_ctr < (sizeof (help_message) / sizeof (char *)); help_ctr ++) /* for each line of help */
            {
                (void) printf ("%s", help_message[help_ctr]); /* print the line */
            }

            break;

        default: /* an error that was not a request for help, print the error */

            (void) fprintf (stderr, "%s", error_message[retval]);
            break;

    }

}

/*

parse a record based on sequential delimiters

int strtoken (char *string, char *parse_array, char **parse, char *delim);

parse a character array, string, into an array, parse_array, using
consecutive characters from delim as field delimiters, point the
character pointers, token, to the beginning of each field, return the
number of fields parsed

*/

#ifdef __STDC__

static int strtoken (char *string, char *parse_array, char **parse, char *delim)

#else

static int strtoken (string, parse_array, parse, delim)
char *string,
*parse_array,
**parse,
*delim;

#endif

{
    int tokens = 0;

    (void) strcpy (parse_array, string); /* copy the string */

    parse[tokens] = strtok (parse_array, delim); /* get the 1st field */

    while (parse[tokens] != 0) /* get the remaining fields */
    {
        parse[++ tokens] = strtok ((char *) 0, delim);
    }

    return (tokens); /* return the number of tokens parsed */
}

/*

double_sort.c - binary sort routine for type double

static void double_sort (double *array, int bottom, int top);

double_sort (), quick sorts an array of type double, and returns
nothing.

note: this is a recursive proceedure, requiring three references, and
4 integers to be allocated on the stack for each recursion

note: on entry, argument "bottom" should be zero, and argument "top"
should be the index of the last array element, ie., if the array
contains n many elements, then the call would be:

    double_sort (array, 0, n - 1);

*/

#ifdef __STDC__

static void double_sort (double *array, int bottom, int top)

#else

static void double_sort (array, bottom, top)
    double *array;
    int bottom;
    int top;

#endif

{
    int i = bottom, /* bottom index of the array */
        j = top; /* top index of the array */

    double x,
           y;

    x = array[MIDWAY(bottom, top)]; /* starting in the middle of the array */

    do /* while the bottom moving index of the array is less than or equal to the top moving index of the array */
    {

        while (array[i] < x && i < top) /* increment the bottom moving index of the array while the elements referenced are lexically "less" */
        {
            i ++;
        }

        while (array[j] > x && j > bottom) /* decrement the top moving index of the array while the elements referenced are lexically "greater" */
        {
            j --;
        }

        if (i <= j) /* if the bottom moving index is less than or equal to the top moving index of the array */
        {
            y = array[i]; /* swap the element references, increment the bottom moving index, decrement the top moving index */
            array[i] = array[j];
            array[j] = y;
            i ++;
            j --;
        }

    }
    while (i <= j);

    if (bottom < j) /* if the top moving index is not at the bottom of the array elements, sort the remaining elements */
    {
        double_sort (array, bottom, j);
    }

    if (i < top) /* if the bottom moving index is not at the top of the array elements, sort the remaining elements */
    {
        double_sort (array, i, top);
    }

}
